#include <math.h>
#include <stdlib.h>
#if defined(WIN32)
#  pragma warning(disable: 4996)
#  include "glew.h"
#  include "glut.h"
#  include "glext.h"
#elif defined(__APPLE__) || defined(MACOSX)
#  include <GLUT/glut.h>
#else
#  define GL_GLEXT_PROTOTYPES
#  include <GL/glut.h>
#endif

/* 頂点バッファオブジェクトのメモリを参照するポインタのデータ型 */
typedef GLfloat Texture[2];
typedef GLuint Face[3];

/*
** メッシュを三角形で描く
*/
GLuint solidMesh(int slices, int stacks, const GLuint *buffer)
{
  GLuint vertices = (slices + 1) * (stacks + 1);
  GLuint faces = slices * stacks * 2;

  /* 頂点バッファオブジェクトを有効にする */
  glBindBuffer(GL_ARRAY_BUFFER, buffer[0]);
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffer[1]);

  /* 頂点バッファオブジェクトにメモリ領域を確保する */
  glBufferData(GL_ARRAY_BUFFER, sizeof (Texture) * vertices, NULL, GL_STATIC_DRAW);
  glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof (Face) * faces, NULL, GL_STATIC_DRAW);

  /* 頂点バッファオブジェクトのメモリをプログラムのメモリ空間にマップする */
  Texture *texture = (Texture *)glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
  Face *face = (Face *)glMapBuffer(GL_ELEMENT_ARRAY_BUFFER, GL_WRITE_ONLY);

  /* 頂点の位置 */
  for (int j = 0; j <= stacks; ++j) {
    float t = (float)j / (float)stacks;

    for (int i = 0; i <= slices; ++i) {
      float s = (float)i / (float)slices;

      (*texture)[0] = s;
      (*texture)[1] = t;
      ++texture;
    }
  }

  /* 面の指標 */
  for (int j = 0; j < stacks; ++j) {
    for (int i = 0; i < slices; ++i) {
      int count = (slices + 1) * j + i;

      /* 上半分 */
      (*face)[0] = count;
      (*face)[1] = count + 1;
      (*face)[2] = count + slices + 2;
      ++face;

      /* 下半分 */
      (*face)[0] = count;
      (*face)[1] = count + slices + 2;
      (*face)[2] = count + slices + 1;
      ++face;
    }
  }

  /* 頂点バッファオブジェクトのメモリをプログラムのメモリ空間から切り離す */
  glUnmapBuffer(GL_ELEMENT_ARRAY_BUFFER);
  glUnmapBuffer(GL_ARRAY_BUFFER);

  /* 頂点バッファオブジェクトを解放する */
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
  glBindBuffer(GL_ARRAY_BUFFER, 0);

  return faces * 3;
}
